#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#include "globals.h"
#include "init.h"
#include "user.h"
#include "end.h"

/*---------------------------------------------*/
/* configfile syntax & conversion description  */
/*---------------------------------------------*/
#define LCC_KEYWORDNUM		4

#define LCC_C_SRV_IP		0
#define LCC_C_SRV_PORT		1

char* lcc_conf_keywords[LCC_KEYWORDNUM] = { "srv_ip", "srv_port",
	"name", "passwd" };
void* lcc_conf_addrs[LCC_KEYWORDNUM]; // runtime init
char* lcc_conf_formats[LCC_KEYWORDNUM] = { "%[0-9.]s", "%d",
	"%s", "%s" };
/*---------------------------------------------*/

char dum[250];

int init_addrs()
{
    int i;
    lcc_conf_addrs[ 0] = dum;
    lcc_conf_addrs[ 1] = &(opts->srv_port);
    lcc_conf_addrs[ 2] = opts->name;
    lcc_conf_addrs[ 3] = opts->pass;
// run a check...
    for (i = 0; i<LCC_KEYWORDNUM; i++)
	if ( !(lcc_conf_addrs[i]) )
	{
	    fprintf(stderr, "Programming fault: init.c:init_addrs(): address not initialized!\n");
	    return(-1);
	}
    return(0); // OK!
}

int init()
{
    static char cfgfilename[256];
    static char line[1024];
    int var;
    FILE* cfgfile;
    char* homedir, *homedir2;
    char* lpos;
//    struct in_addr inaddr;
    if ( init_addrs() ) return(-1);
    strcpy(cfgfilename, LCC_CFGFILE);
    printf("checking for ./%s... ", cfgfilename);
    fflush(stdout);
    if ( !(cfgfile = fopen(cfgfilename, "r")) )
    {
	printf("not found.\n");
	homedir = get_homedir(getuid());
	if ( !homedir )
	{
	    printf("unable to determine users home directory.\n");
	    return(-1);
	}
	if ( strlen(homedir) > 256 - strlen(cfgfilename) - 2 )
	{
	    printf("maximum stringlength to small to save possible path to cfgfile!\n");
	    return(-1);
	}
	strcat(homedir, "/");
	homedir2 = strdup(homedir);
	strcat(homedir, cfgfilename);
	strcpy(cfgfilename, homedir);
	printf("checking for %s... ", cfgfilename);
	fflush(stdout);
	if ( !(cfgfile = fopen(cfgfilename, "r")) )
	{
	    printf("not found.\n");
	    strcpy(cfgfilename, homedir2);
	    strcat(cfgfilename, ".");
	    strcat(cfgfilename, LCC_CFGFILE);
	    printf("checking for %s...", cfgfilename);
	    if ( !(cfgfile = fopen(cfgfilename, "r")) )
	    {
		printf("checking for %s... ", LCC_SYSCFGFILE);
		if ( !(cfgfile = fopen(LCC_SYSCFGFILE, "r")) )
		{
		    printf("no configuration file found!\n");
		    free(homedir2);
		    free(homedir);
		    return(-1);
		}
	    }
	}
	free(homedir2);
	free(homedir);
    }
    printf("parsing... ");
    // analize configfile
    while ( fgets(line, sizeof(line), cfgfile) != NULL )
    {
	if ( line[0] == '#' || line [0] == '\n' ) continue;
	for ( var=0; var < LCC_KEYWORDNUM; var++ )
	{
	    if ( strstr(line, lcc_conf_keywords[var]) != line ) continue;
	    lpos = line + strlen(lcc_conf_keywords[var]);
	    if ( *lpos++ != ' ' ) continue;	// have space after each keyword
	    sscanf(lpos, lcc_conf_formats[var], lcc_conf_addrs[var]);
	    switch (var)
	    {
		case LCC_C_SRV_IP:
		    opts->srv_ip = inet_addr(dum);
		    break;
		case LCC_C_SRV_PORT:
		    opts->srv_port = htons(opts->srv_port);
		    break;
	    }
	}
    }
    fclose(cfgfile);
    printf("done.\n");
    // dump a part of the configs
/*    inaddr.s_addr = opts->srv_ip;
    printf("------\n");
    printf("server address: %s:%d\n", inet_ntoa(inaddr), ntohs(opts->srv_port));
    printf("username: %s\n", opts->name);
    printf("------\n");*/
    // dump finished
/*    if ( !(opts->pass[0]) )
    {
	printf("there was no password (keyword: passwd) in the configfile!\n");
	return(-1);
    }*/
    // redirect signals
    signal(SIGINT, cleanup);
    signal(SIGQUIT, cleanup);
    signal(SIGTERM, cleanup);
    return(0); // ok!
}

void get_opts(int argc, char **argv)
{
    int i, c = 0;
    char sbuf[20];
    char *eptr = NULL;
    char *bufidx = sbuf+1;
    char *badparam = "Bad parameter for argument";
    char *fdis = "Taking defaults";
    sbuf[0]=' ';
    for (i = 1; (i < argc) && (c != -1); i++)
    {
	c = getopt(argc, argv, "s:p:n:w:u:d:i:t:xl:g");
	switch ( c )
	{
	    case 's': // server
		if ( strlen(optarg) <= 15 )
		{
		    opts->srv_ip = inet_addr(optarg);
		}
		break;
	    case 'p': // port
		if ( strlen(optarg) <= sizeof(sbuf)-2 )
		{
		    strcpy(bufidx, optarg);
		    opts->srv_port = htons(strtol(sbuf, &eptr, 0));
		    if ( !eptr ) // bad argument
		    {
			//opts->intervals = 3; // already preloaded
			printf("%s -p. %s\n", badparam, fdis);
		    }
		}
		break;
	    case 'n': // username
		if ( strlen(optarg) <= LCP3_MAX_NAME_LEN )
		{
		    strncpy(opts->name, optarg, LCP3_MAX_NAME_LEN+1);
		}
		else
		    printf("Supplied username to long (max %d characters).\n", LCP3_MAX_NAME_LEN);
		break;
	    case 'w': // password
		if ( strlen(optarg) <= LCP3_MAX_PWD_LEN )
		{
		    strncpy(opts->pass, optarg, LCP3_MAX_PWD_LEN+1);
		}
		else
		    printf("Supplied password to long (max %d characters).\n", LCP3_MAX_PWD_LEN);
		break;
	    case 'u': // upload
		if ( strlen(optarg) <= sizeof(sbuf)-2 )
		{
		    strcpy(bufidx, optarg);
		    opts->min_up = strtol(sbuf, &eptr, 0);
		    if ( !eptr ) // bad argument
		    {
			//opts->intervals = 3; // already preloaded
			printf("%s -u. %s\n", badparam, fdis);
		    }
		    else
		    {
			opts->watch_up = 1;
			printf("minimal upstream rate:   %6d B/s\n", (int)opts->min_up);
		    }
		}
		break;
	    case 'd': // download
		if ( strlen(optarg) <= sizeof(sbuf)-2 )
		{
		    strcpy(bufidx, optarg);
		    opts->min_dn = strtol(sbuf, &eptr, 0);
		    if ( !eptr ) // bad argument
		    {
			//opts->intervals = 3; // already preloaded
			printf("%s -d. %s\n", badparam, fdis);
		    }
		    else
		    {
			opts->watch_dn = 1;
			printf("minimal downstream rate: %6d B/s\n", (int)opts->min_dn);
		    }
		}
		break;
	    case 'i': // interval
		if ( strlen(optarg) <= sizeof(sbuf)-2 )
		{
		    strcpy(bufidx, optarg);
		    opts->intervals = strtol(sbuf, &eptr, 0);
		    if ( !eptr ) // bad argument
		    {
			//opts->intervals = 3; // already preloaded
			printf("%s -i. Using %d intervals.\n", badparam, (int)opts->intervals);
		    }
		    else printf("up/downstream messages to check: %d\n", (int)opts->intervals);
		}
		break;
	    case 't': // timed
		if ( strlen(optarg) <= sizeof(sbuf)-2 )
		{
		    strcpy(bufidx, optarg);
		    opts->timed = strtol(sbuf, &eptr, 0);
		    if ( !eptr ) // bad argument
		    {
			//opts->intervals = 3; // already preloaded
			printf("%s -t. %s\n", badparam, fdis);
		    }
		    else
		    {
			opts->istimed = 1;
			printf("Online time limited to %d sec.\n", (int)opts->timed);
		    }
		}
		break;
	    case 'x': // stay offline
		opts->stay_offline = 1;
		break;
	    case 'l':
		if ( strlen(optarg) <= sizeof(sbuf)-2 )
		{
		    strcpy(bufidx, optarg);
		    opts->line_id = strtol(sbuf, &eptr, 0);
		    if ( !eptr ) // bad argument
		    {
			opts->line_id = 1;
			printf("%s -l. %s\n", badparam, fdis);
		    }
		    else
		    {
			printf("Line %d selected.\n", (int)opts->line_id);
		    }
		}
		break;
	    case 'g':
		opts->get_lines = 1;
		break;
	}
    }
}
