#include <stdio.h>
#include <time.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <stdlib.h>
#include "lcp3_io.h"
#include "lcp3.h"
#include "lists.h"
#include "globals.h"
#include "end.h"

struct t_queue_ent {
    struct list_hdr_t hdr;
    int cmd;
    void *info;
    int isize;
    time_t prev_try;
    int cnt;
};

int lcp3_socket;
struct list_anchor_t cmd_q;


void proc_queue();

//----- end of decs ----


void lcp3_io_init(int socket)
{
    lcp3_socket = socket;
    list_init(&cmd_q);
}

void cmd_send(int cmd, void *info, int isize)
{
    int slen = 0, plen = 0;
    struct t_lcp3_cmd* po_cmd = outbuf;
    void* po_info = ((char*)outbuf) + sizeof(struct t_lcp3_cmd);
	fd_set wfd;
	struct timeval tv;
    po_cmd->magic = LCP3_MAGICNR;
    po_cmd->version = LCP3_PROTO_VERSION;
    po_cmd->cmd = cmd;
    po_cmd->line_id = opts->line_id; // default line
    if ( info ) memcpy(po_info, info, isize);
    plen = sizeof(struct t_lcp3_cmd) + isize;
//fprintf(stderr, "sending cmd %d...\n", cmd);
	tv.tv_sec = 2;
	tv.tv_usec = 0;
	FD_ZERO(&wfd);
	FD_SET(lcp3_socket, &wfd);
	if ( !select(lcp3_socket+1, NULL, &wfd, NULL, &tv) )
		fprintf(stderr, "select() for writing failed! can't send command %d.\n", cmd);
	else
    	slen = send(lcp3_socket, outbuf, plen, 0);
	if ( slen != plen )
		fprintf(stderr, "send() of command %d failed.\n", cmd);
    globs->prevsent = time(NULL);
}

void cmd_queue(int cmd, void *info, int isize)
{
    struct t_queue_ent* newbie;
    if ( (sizeof(struct t_lcp3_cmd) + isize) != get_lcp3_pack_size(cmd) )
    {
	fprintf(stderr, "lcp3 error: bad cmd size: cmd %d should be %d bytes instead of %d!\n",
	    cmd, get_lcp3_pack_size(cmd), isize + sizeof(struct t_lcp3_cmd));
	return;
    }
    newbie = (struct t_queue_ent*)list_add(&cmd_q, sizeof(struct t_queue_ent));
    if ( !newbie )
    {
	fprintf(stderr, "not enough memory to put a command in the queue!\n");
	close(lcp3_socket);
	exit(-6);
    }
    newbie->cmd = cmd;
    newbie->info = info;
    newbie->isize = isize;
    newbie->cnt = 0;
    newbie->prev_try = 0;
    proc_queue();
}

void clean_queue()
{
    void* dum;
    struct t_queue_ent* cmd = (struct t_queue_ent*)cmd_q.first;
    while ( cmd )
    {
	if ( (cmd->prev_try) && (cmd->prev_try < time(NULL) - LCP3_TMO_ACK) && (cmd->cnt < 3) )
	{ // resend
	    cmd_send(cmd->cmd, cmd->info, cmd->isize);
	    cmd->cnt++;
	    cmd->prev_try = time(NULL);
	    break;
//	    cmd = (struct t_queue_ent*)cmd->hdr.next;
	}
	else if ( cmd->cnt >= 3 )// delete command
	{
	    dum = cmd->hdr.next;
	    free(cmd->info);
	    list_del(&cmd_q, (struct list_hdr_t*)cmd);
	    cmd = dum;
	}
	else
	    cmd = (struct t_queue_ent*)cmd->hdr.next;
    }
}

void proc_queue()
{
    struct t_queue_ent* cmd;
    clean_queue();
    // first check whether we're waiting for an ack
    cmd = (struct t_queue_ent*)cmd_q.first;
    while ( cmd )
    {
	if ( cmd->cnt ) return;
	cmd = (struct t_queue_ent*)cmd->hdr.next;
    }
    // send an unsent command
    cmd = (struct t_queue_ent*)cmd_q.first;
    while ( cmd )
    {
	if ( !(cmd->cnt) )
	{
	    cmd_send(cmd->cmd, cmd->info, cmd->isize);
	    cmd->cnt++;
	    cmd->prev_try = time(NULL);
	    break;
	}
	cmd = (struct t_queue_ent*)cmd->hdr.next;
    }
}

int get_data()
// returns 0 if we have to process the buffer "packet"
// -1 on error or when there is not enough data
{
    fd_set rfd;
    struct timeval tv;
    int nd, size, i;
    int available = get_max_lcp3_pack_size() * 3 - globs->received;
    char* ptr = inbuf + globs->received;
    char* ptb;
    unsigned int* puint = NULL;
    struct t_lcp3_cmd *pcmd = (struct t_lcp3_cmd *)inbuf;
    tv.tv_sec = 0;
    tv.tv_usec = 0;
    FD_ZERO(&rfd);
    FD_SET(lcp3_socket, &rfd);
    if ( select(lcp3_socket+1, &rfd, NULL, NULL, &tv) == 1 )
    {
	nd = read(lcp3_socket, ptr, available);
	if ( nd != -1 ) globs->received += nd;
    }
    // seek magicnr
    ptr = inbuf;
    ptb = inbuf;
    puint = (int*)ptr;
    for ( i = 0; i < globs->received - 3; i++)
    {
	if ( *puint == LCP3_MAGICNR )
	    break;
	ptr++;
	puint = (int*)ptr;
    }
    // either we have to clean the buffer or ptr points to magicnr
    globs->received -= i; // trash i bytes
    for (i = 0; i < globs->received; i++)
    {
	*ptb = *ptr;
	ptb++;
	ptr++;
    }
    // check for a valid packet
    if ( globs->received < sizeof(struct t_lcp3_cmd) ) return(-1);
    if ( pcmd->version != LCP3_PROTO_VERSION )
    {
	fprintf(stdout, "Server uses bad protocol version!!\n");
	for (;;) { cleanup(); sleep(1); }
	// prog won't run until here!
    }
    size = get_lcp3_pack_size(pcmd->cmd);
    if ( globs->received < size ) return(-1);
    // packet should be ok, copy it
    memcpy(packet, inbuf, size);
    // move the rest of the data to the beginning of the buffer
    globs->received -= size;
    ptr = inbuf;
    ptb = inbuf + size;
    for ( i = globs->received; i; i-- )
    {
	*ptr = *ptb;
	ptr++;
	ptb++;
    }
    if ( globs->received < 0 )
    {
	fprintf(stderr, "get_data(): The programmer is a fool!\n");
	globs->received = 0;
    }
/*
    // check whether we have a complete packet
    if ( globs->received < sizeof(struct t_lcp3_cmd) ) return(-1);
    size = get_lcp3_pack_size(((struct t_lcp3_cmd*)inbuf)->cmd);
    if ( globs->received < size ) return(-1);
    // copy the complete packet
    memcpy(packet, inbuf, size);
    globs->received -= size;
//fprintf(stderr, "got cmd %d; %d bytes left in input buffer.\n",
//    ((struct t_lcp3_cmd*)inbuf)->cmd, globs->received);
    // move the remaining data to the beginning of the inbuf
    ptb = inbuf;
    ptr = inbuf + size;
    // seek the next magicnr (first int of a cmd struct)
    for ( i = 0; i < globs->received; ptr++, puint = (int*)ptr, i++)
	if ( *puint == LCP3_MAGICNR )
	    break;
    size = i;
    // ptb: inbuf
    // ptr: beginning of possibly valid packet
    // size: bad bytes to jump over
    globs->received -= size;
    // copy the rest of the packet to the beginning of the buffer
    for (i = 0; i < globs->received; i++)
    {
	*ptb = *ptr;
	ptb++;
	ptr++;
    }
*/
    return(0);
}

void proc_ack(struct t_lcp3_cmd* pcmd)
{
    struct t_queue_ent* cmd = (struct t_queue_ent*)cmd_q.first;
    struct t_lcp3_info_ack* ack = (struct t_lcp3_info_ack*)((char*)pcmd + sizeof(struct t_lcp3_cmd));
//fprintf(stderr, "proc ack... ");
    while ( cmd )
    {
	if ( (ack->ackcmd == cmd->cmd) && cmd->prev_try ) break;
	cmd = (struct t_queue_ent*)cmd->hdr.next;
    }
    // delete the cmd if one was found
    if ( cmd )
    {
//fprintf(stderr, "ack was for %d\n", cmd->cmd);
	free(cmd->info);
	list_del(&cmd_q, (struct list_hdr_t*)cmd);
    }
    // check whether we have to send a command
    proc_queue();
}

void send_ack(int cmd)
// sends an ack for cmd
{
    static struct t_lcp3_info_ack iack;
    iack.ackcmd = cmd;
//fprintf(stderr, "sending ack for %d\n", cmd);
    cmd_send(CMD3_ACK, &iack, sizeof(struct t_lcp3_info_ack));
}
