#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>

#include "globals.h"
#include "lcp3.h"
#include "lcp3_io.h"
#include "end.h"

void make_time(char *tstr, int size, time_t secs)
{
    char *idx;
    snprintf(tstr, size, "%d:", (int)(secs / 3600)); // print hours
    secs %= 3600;
    idx = tstr + strlen(tstr);
    if ( secs / 60 < 10 ) { idx[0] = '0'; idx++; } // eventually print 0
    snprintf(idx, size-strlen(tstr), "%d:", (int)(secs / 60)); // print minutes
    secs %= 60;
    idx = tstr + strlen(tstr);
    if ( secs < 10 ) { idx[0] = '0'; idx++; }
    snprintf(idx, size-strlen(tstr), "%d", (int)(secs));
}


void update_line_status(struct t_lcp3_cmd* lcp3)
{
	    if ( (!lcp3->info_is_valid) || (lcp3->line_id != opts->line) ) return;
	    if ( globs->line == lcp3->line_stat ) return;
	    globs->line = lcp3->line_stat;
	    if ( opts->get_lines ) return;
	    printf("LINE %d: ", lcp3->line_id);
	    switch ( lcp3->line_stat )
	    {
		case CST_DOWN: printf("closed"); break;
		case CST_DIALING: printf("dialing..."); break;
		case CST_UP_SERVER: printf("established"); break;
		case CST_UP_USER: printf("manually estab."); break;
		case CST_CLOSING: printf("closing..."); break;
		case CST_CLOSE_ERROR: printf("close error"); break;
		case CST_LOCKED_UP_CLT: printf("established & locked"); break;
		case CST_LOCKED_UP_TIMED: printf("established, timer"); break;
		case CST_LOCKED_DOWN: printf("disabled"); break;
	    }
	    printf("   CHANNELS: %d\n\n", lcp3->line_channels);
}

int process_cmd(struct t_lcp3_cmd* lcp3)
{
#ifndef OSD
    static char strbuf[16];
#else
#define OUTBUFLEN 200
    static char outbuf[OUTBUFLEN];
#endif
    struct t_lcp3_info_enckey *info_enckey;// appended to a CMD3_SRVPUBKEY
    struct t_lcp3_info_ack *info_ack;	// appended to a CMD3_ACK
    struct t_lcp3_info_line *info_line;	// appended to a CMD3_LINE
    struct t_lcp3_info_tput *info_tput;	// appended to a CBR3_TPUT
    struct t_lcp3_info_status *info_status;// extended line status
    struct t_lcp3_info_isdn *info_isdn;
    void *info_ptr;
    void *ptr = ((char*)lcp3)+sizeof(struct t_lcp3_cmd);
//	time_t intv;
    info_enckey = ptr;
    info_ack = ptr;
    info_line = ptr;
    info_tput = ptr;
    info_status = ptr;
    info_isdn = ptr;
/*
printf("got %d ", lcp3->cmd);
if ( lcp3->cmd == CMD3_ACK )
    printf("for %d\n", info_ack->ackcmd);
else if ( lcp3->cmd == CMD3_CLIENTSTAT )
    printf("status %d\n", lcp3->client_stat);
else printf("\n");
*/
    // send an ack ?
    if ( (lcp3->cmd < MIN_CBR3) && (lcp3->cmd != CMD3_ACK) )
	send_ack(lcp3->cmd);
    // line status update?
    if ( lcp3->info_is_valid && (opts->line == lcp3->line_id)
	    && (globs->line != lcp3->line_stat) )
	update_line_status(lcp3);
    // status update?
    if ( lcp3->info_is_valid && (globs->client != lcp3->client_stat) )
    {
	globs->client = lcp3->client_stat;
	if ( opts->get_lines ) 
	{
	    if ( lcp3->client_stat == CLT_OFFLINE )
	    {
		cmd_queue(CMD3_LINESREQ, NULL, 0);
	    }
	}
	else
	{
	    printf("CLIENT: ");
	    switch ( lcp3->client_stat )
	    {
	        case CLT_DISCON:
		    printf("disconneted from server\n\n");
		    cleanup();
		    break;
		case CLT_NOAUTH: //printf("connected to srv, not yet authenticated\n\n");
		    printf("Not authenticated!\n");
		    break;
		case CLT_OFFLINE:
		    printf("OFFLINE\n\n");
		    switch ( globs->starting )
		    {
			case 0: cleanup(); break;
			case 1:
			    cmd_queue(CMD3_LINESEL, NULL, 0);
			    if ( opts->stay_offline || opts->get_lines )
			    {
				if ( opts->stay_offline )
				    cmd_queue(CMD3_STATREQ, NULL, 0);
				else if ( opts->get_lines )
				    cmd_queue(CMD3_LINESREQ, NULL, 0);
			    }
			    else
			    {
				switch ( globs->line )
				{
				    case CST_DOWN:
				    case CST_DIALING:
				    case CST_UP_SERVER:
					cmd_queue(CMD3_ONLINE, NULL, 0);
					break;
				    default: cleanup();
				}
			    }
			    break;
		    }
		    break;
	        case CLT_DIALING: printf("dialing...\n\n"); break;
	        case CLT_ONLINE:
		    printf("ONLINE\n\n");
		    globs->starting = 0;
		    break;
	        case CLT_CLOSING: printf("closing line...\n\n"); break;
	    }
	}
    }
    // handle command
    switch ( lcp3->cmd )
    {
	case CMD3_AUTHBAD: // TCP connection gets dropped by server
	    printf("failed.\n\n");
	    globs->client = CLT_NOAUTH;
	    cleanup();
	    break;
	case CMD3_CLIENTSTAT:
	    break;
	case CMD3_LINE: // "struct t_lcp3_info_line" appended
	    printf("line %d: '%s'\n", info_line->line_id, info_line->linename);
	    if ( (info_line->line_id == info_line->total_lines)
		&& (opts->get_lines) )
	    {
		printf("\n");
		cleanup();
	    }
	    break;
	case CMD3_STATUS: // "struct t_lcp3_info_status" appended
	    if ( ! (lcp3->info_is_valid) ) break;
	    printf("Serverstatus:\n clients: %d\n online on line %d: %d\n\n",
		    info_status->clients, lcp3->line_id, info_status->online);
	    break;
	case CMD3_INVALID: // "struct t_lcp3_info_ack" appended
	    switch ( info_ack->ackcmd )
	    {
		case CMD3_SHUTDOWN:
		    printf("Shutdown feature is not implemented in server.\n");
		    cleanup();
		    break;
		default:
		    printf("Server said that command %d is invalid!\n", info_ack->ackcmd);
	    }
	    break;
	case CMD3_ENCINFO: // "struct t_lcp3_info_enckey" appended
	    if ( !(info_ptr = malloc(sizeof(struct t_lcp3_info_auth))) )
	    {
		fprintf(stderr, "Out of memory...\n");
		cleanup(0);
	    }
	    printf("Preferred authentication method: ");
	    switch ( info_enckey->enctype )
	    {
		case LCP3_ENCTYPE_NONE:
		    printf("plain text\n");
		    break;
		case LCP3_ENCTYPE_SYM:
		    printf("symmetric encryption\n");
		    break;
		case LCP3_ENCTYPE_ASYM:
		    printf("asymmetric encryption\n");
		    break;
		case LCP3_ENCTYPE_OPEN:
		    printf("no authentication required\n");
		    break;
	    }
	    memset(info_ptr, 0, sizeof(struct t_lcp3_info_auth));
	    printf("Used authentication method:      ");
	    switch ( info_enckey->enctype )
	    {
/*		case LCP3_ENCTYPE_NONE:
		case LCP3_ENCTYPE_SYM:
		case LCP3_ENCTYPE_ASYM:
		    break;*/
		case LCP3_ENCTYPE_OPEN:
		    printf("no authentication\n");
		    // don't supply name / pwd
		    ((struct t_lcp3_info_auth*)info_ptr)->enctype = LCP3_ENCTYPE_OPEN;
		    ((struct t_lcp3_info_auth*)info_ptr)->name[0] = 0;
		    ((struct t_lcp3_info_auth*)info_ptr)->pass[0] = 0;
		    break;
		default:
		    printf("plain text\n");
		    printf("username: %s\n", opts->name);
		    // use LCP3_ENCTYPE_NONE
		    ((struct t_lcp3_info_auth*)info_ptr)->enctype = LCP3_ENCTYPE_NONE;
		    strncpy(((struct t_lcp3_info_auth*)info_ptr)->name, opts->name, LCP3_MAX_NAME_LEN);
//		    ((struct t_lcp3_info_auth*)info_ptr)->name[LCP3_MAX_NAME_LEN] = 0;
		    strncpy(((struct t_lcp3_info_auth*)info_ptr)->pass, opts->pass, LCP3_MAX_PWD_LEN);
//		    ((struct t_lcp3_info_auth*)info_ptr)->pass[LCP3_MAX_PWD_LEN] = 0;
		    break;
	    }
	    printf("\n");
	    cmd_queue(CMD3_AUTHREQ, info_ptr, sizeof(struct t_lcp3_info_auth));
	    break;
	case CMD3_NOACCESS: // "struct t_lcp3_info_ack" appended
	    switch ( info_ack->ackcmd )
	    {
		case CMD3_SHUTDOWN:
		    printf("You're not allowed to shutdown the server!\n");
		    break;
		default:
		    printf("You have no access to command %d\n", info_ack->ackcmd);
	    }
	    break;
	case CMD3_CHANFAIL:
	    printf("Manipulating ISDN channels failed.\n");
	    break;
	case CMD3_ACK:
	    proc_ack(lcp3);
	    break;
	case CMD3_UNREG:
	    printf("Unregistered, disconnected from server.\n\n");
	    close(globs->socket);
	    globs->client = CLT_DISCON;
	    cleanup();
	    break;
	case CBR3_TPUT: // "struct t_lcp3_info_tput" appended
		if ( opts->get_lines ) break;
#ifndef OSD
		if ( globs->first_tput )
		{
		    printf("download                   upload                     time\n");
		    printf("B/s     avg     MB total   B/s     avg     MB total   line       client\n");
		    printf("-------------------------  -------------------------  -------------------\n");
		    globs->first_tput = 0;
		}
		printf("\r%6d, %6d, %9d  %6d, %6d, %9d  ",
			info_tput->dns, info_tput->dnl, info_tput->dnt / 1024,
			info_tput->ups, info_tput->upl, info_tput->upt / 1024);
		make_time(strbuf, sizeof(strbuf), info_tput->line);
		printf("%8s   ", strbuf);
		make_time(strbuf, sizeof(strbuf), info_tput->client);
		printf("%8s   ", strbuf);
#else
		intv = time(NULL) - osd.line2_active_since;
		if ( !((intv >= 0) && (intv < 20)) )
			snprintf(outbuf, OUTBUFLEN, "     %6d B/s up,   %6d B/s down", info_tput->ups, info_tput->dns);
		xosd_display(osd.osd, 0, XOSD_string, outbuf);
#endif
	    /* now check the limits... */
	    if ( opts->watch_up && (opts->min_up > abs(info_tput->ups)) )
		globs->limc_up++;
	    else globs->limc_up = 0;
	    if ( opts->watch_dn && (opts->min_dn > abs(info_tput->dns)) )
		globs->limc_dn++;
	    else globs->limc_dn = 0;
/*	    if ( globs->autohup_and )
	    {
		if (  ((io_conf.limc_up >= opts.tput_intv)
		       && (io_conf.limc_down >= opts.tput_intv))
		     && (opts.tput_intv) )
		{
		    if ( !opts.quiet ) printf("traffic lower than both of the two possible limits.\n\n");
		    cleanup(0);
		}
	    }
	    else*/
	    {
		if (  ((globs->limc_up >= opts->intervals)
		       || (globs->limc_dn >= opts->intervals))
		     && (opts->intervals) )
		{
		    printf("traffic lower than one of the two possible limits.\n\n");
		    cmd_queue(CMD3_OFFLINE, NULL, 0);
		}
	    }
	    if ( opts->istimed && (opts->timed < info_tput->client)
		 && (globs->client == CLT_ONLINE) )
	    {
		printf("time is over.\n\n");
		cmd_queue(CMD3_OFFLINE, NULL, 0);
	    }
	    break;
	case CBR3_LINESTAT: // has been done at the beginning
//	    update_line_status(lcp3);
	    break;
	case CBR3_SRVDEATH:
	    printf("Server terminated.\n\n");
	    close(globs->socket);
	    globs->client = CLT_DISCON;
	    cleanup();
	    break;
	case CBR3_ISDNACT: // "struct t_lcp3_info_isdn" appended
	#ifndef OSD
	    printf("Incoming ISDN call:\n caller: %s\n call to: %s\n time: %s %s.%s.%s\n\n",
		    info_isdn->caller, info_isdn->on, info_isdn->time,
		    info_isdn->day, info_isdn->month, info_isdn->year);
	#else
		snprintf(outbuf, OUTBUFLEN, "call from %s", info_isdn->caller);
		xosd_display(osd.osd, 0, XOSD_string, outbuf);
		osd.line2_active_since = time(NULL);
	#endif
	    break;
    }
    fflush(stdout);
    return(0);
}
